package org.processmining.glueminer.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.glueminer.models.OrderedActivity;
import org.processmining.glueminer.models.OrderedActivityComparator;
import org.processmining.glueminer.models.PlaceUsageStatus;

public class PlaceSearchRecursion {
	Double placeFitnessLevel;
	Double placeUnderfednessLevel;
	PlaceAdditionTriager placeAdditionTriager;
	Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheLeftSide;
	Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheRightSide;
	Map<OrderedActivity, Set<OrderedActivity>> possibleActivitiesOnTheRightMap;
	List<OrderedActivity> orderedActivitiesToExamineOnTheLeftSide;
	Set<OrderedActivity> allActivitiesToConsiderOnTheRightSide;
	Integer maxSourceOrTargetTransitions;
	Integer minSupportForPlaces;
	List<Double> fitPlacesButNotSoFit;
	List<Double> fitButAlsoUnderfed;
	
	public PlaceSearchRecursion(Double placeFitnessLevel, Double placeUnderfednessLevel, PlaceAdditionTriager placeAdditionTriager, Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheLeftSide, Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheRightSide, Map<OrderedActivity, Set<OrderedActivity>> possibleActivitiesOnTheRightMap, Integer maxSourceOrTargetTransitions, Integer minSupportForPlaces) {
		this.placeFitnessLevel = placeFitnessLevel;
		this.placeUnderfednessLevel = placeUnderfednessLevel;
		this.placeAdditionTriager = placeAdditionTriager;
		this.nonImportantRelationshipsOnTheLeftSide = nonImportantRelationshipsOnTheLeftSide;
		this.nonImportantRelationshipsOnTheRightSide = nonImportantRelationshipsOnTheRightSide;
		this.possibleActivitiesOnTheRightMap = possibleActivitiesOnTheRightMap;
		
		this.orderedActivitiesToExamineOnTheLeftSide = new ArrayList<OrderedActivity>(this.possibleActivitiesOnTheRightMap.keySet());
		Collections.sort(this.orderedActivitiesToExamineOnTheLeftSide, new OrderedActivityComparator());
		
		this.allActivitiesToConsiderOnTheRightSide = new HashSet<OrderedActivity>();
		for (OrderedActivity activity : possibleActivitiesOnTheRightMap.keySet()) {
			this.allActivitiesToConsiderOnTheRightSide.addAll(possibleActivitiesOnTheRightMap.get(activity));
		}
		
		this.maxSourceOrTargetTransitions = maxSourceOrTargetTransitions;
		this.minSupportForPlaces = minSupportForPlaces;
		
		this.fitPlacesButNotSoFit = new ArrayList<Double>();
		this.fitButAlsoUnderfed = new ArrayList<Double>();
	}
	
	public void startRecursion() {
		Set<OrderedActivity> addedLeftActivities = new HashSet<OrderedActivity>();
		Set<OrderedActivity> addedRightActivities = new HashSet<OrderedActivity>();
		
		recursionFunction(addedLeftActivities, addedRightActivities, orderedActivitiesToExamineOnTheLeftSide, allActivitiesToConsiderOnTheRightSide, PlaceUsageStatus.status.NOTKNOWN);
		Collections.sort(fitPlacesButNotSoFit);
		Collections.reverse(fitPlacesButNotSoFit);
		Collections.sort(fitButAlsoUnderfed);
		
		System.out.println("fitPlacesButNotSoFit=");
		System.out.println(fitPlacesButNotSoFit);
		System.out.println("fitButAlsoUnderfed=");
		System.out.println(fitButAlsoUnderfed);
	}
	
	public Integer sumOrderedActivityOccurrences(Set<OrderedActivity> activities) {
		Integer sum = 0;
		
		for (OrderedActivity activity : activities) {
			sum += activity.occurrences;
		}
		
		return sum;
	}
	
	public void recursionFunction(Set<OrderedActivity> addedLeftActivities, Set<OrderedActivity> addedRightActivities, List<OrderedActivity> leftActivitiesRemainingToBeAdded, Set<OrderedActivity> rightActivitiesRemainingToBeAdded, PlaceUsageStatus.status placeUsageStatus) {	
		// helps to reduce computation time
		double placeUnderfedness = 0.0;
		
		if (addedLeftActivities.size() > 0 && addedRightActivities.size() > 0) {
			if (addedLeftActivities.size() <= this.maxSourceOrTargetTransitions && addedRightActivities.size() <= this.maxSourceOrTargetTransitions) {
				Set<OrderedActivity> checkIntersection = new HashSet<OrderedActivity>(addedLeftActivities);
				checkIntersection.retainAll(addedRightActivities);
				
				if (checkIntersection.size() == 0) {
					if (!placeAdditionTriager.checkRedundant(addedLeftActivities, addedRightActivities)) {
						Integer sumOfAddedLeftActivities = sumOrderedActivityOccurrences(addedLeftActivities);
						Integer sumOfAddedRightActivities = sumOrderedActivityOccurrences(addedRightActivities);
						Integer minOfSums = Math.min(sumOfAddedLeftActivities, sumOfAddedRightActivities);
						if (minOfSums > this.minSupportForPlaces) {
							//System.out.println("trying adding place addedLeftActivities = "+addedLeftActivities+" addedRightActivities="+addedRightActivities);
							Object[] placeRet = placeAdditionTriager.tryAddPlace(addedLeftActivities, addedRightActivities, placeFitnessLevel, placeUnderfednessLevel);
							placeUsageStatus = (PlaceUsageStatus.status) placeRet[0];
							double placeFitness = (double) placeRet[1];
							placeUnderfedness = (double) placeRet[2];
							
							if (placeUsageStatus == PlaceUsageStatus.status.FIT) {
								//System.out.println("placeStatus = "+placeUsageStatus+" placeFitness="+placeFitness);
								if (placeFitness < placeFitnessLevel) {
									fitPlacesButNotSoFit.add(placeFitness);
								}
								else if (placeUnderfedness >= placeUnderfednessLevel) {
									fitButAlsoUnderfed.add(placeUnderfedness);
								}
							}
						}
					}
				}
			}
		}
		
		if (addedLeftActivities.size() == 0 || placeUsageStatus.equals(PlaceUsageStatus.status.UNDERFED) || addedRightActivities.size() > this.maxSourceOrTargetTransitions) {
			// we must add at least a left activity
			for (OrderedActivity activity : leftActivitiesRemainingToBeAdded) {
				Set<OrderedActivity> newAddedLeftActivities = new HashSet<OrderedActivity>(addedLeftActivities);
				newAddedLeftActivities.add(activity);
				Integer activityIndex = leftActivitiesRemainingToBeAdded.indexOf(activity);
				List<OrderedActivity> newLeftActivitiesRemaining = new ArrayList<OrderedActivity>(leftActivitiesRemainingToBeAdded);
				newLeftActivitiesRemaining = newLeftActivitiesRemaining.subList(activityIndex+1, newLeftActivitiesRemaining.size());
				
				if (nonImportantRelationshipsOnTheLeftSide.containsKey(activity)) {
					newLeftActivitiesRemaining.retainAll(nonImportantRelationshipsOnTheLeftSide.get(activity));
				}
				recursionFunction(newAddedLeftActivities, addedRightActivities, newLeftActivitiesRemaining, rightActivitiesRemainingToBeAdded, placeUsageStatus);
			}
		}
		else if (addedRightActivities.size() == 0 || placeUsageStatus.equals(PlaceUsageStatus.status.OVERFED)) {
			if (placeUnderfedness <= placeUnderfednessLevel) {
				// we must add at least a right activity
				for (OrderedActivity activity : rightActivitiesRemainingToBeAdded) {
					Set<OrderedActivity> newAddedRightActivities = new HashSet<OrderedActivity>(addedRightActivities);
					newAddedRightActivities.add(activity);
					Set<OrderedActivity> newRightActivitiesRemaining = new HashSet<OrderedActivity>(rightActivitiesRemainingToBeAdded);
					newRightActivitiesRemaining.remove(activity);
					
					if (nonImportantRelationshipsOnTheRightSide.containsKey(activity)) {
						newRightActivitiesRemaining.retainAll(nonImportantRelationshipsOnTheRightSide.get(activity));
					}
					recursionFunction(addedLeftActivities, newAddedRightActivities, leftActivitiesRemainingToBeAdded, newRightActivitiesRemaining, placeUsageStatus);
				}
			}
		}
	}
}
