package org.processmining.glueminer.models;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class IndirectlyFollowsMatrix {
	Map<String, Map<String, Integer>> indirectlyFollowsMatrix;
	Map<String, Integer> maxSuccessOcc;
	Map<String, Map<String, Integer>> invIndirectlyFollowsMatrix;
	Map<String, Integer> maxPrecedOcc;
	Map<String, Map<String, Integer>> importantIndirectlyFollowsMatrix;
	Map<String, Map<String, Integer>> importantInvIndirectlyFollowsMatrix;
	Map<String, Map<String, Integer>> reallyImportantIndirectlyFollowsMatrix;
	Map<String, Map<String, Integer>> reallyImportantInvIndirectlyFollowsMatrix;
	
	public IndirectlyFollowsMatrix() {
		indirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
		invIndirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
		maxSuccessOcc = new HashMap<String, Integer>();
		maxPrecedOcc = new HashMap<String, Integer>();
		importantIndirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
		importantInvIndirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
		reallyImportantIndirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
		reallyImportantInvIndirectlyFollowsMatrix = new HashMap<String, Map<String, Integer>>();
	}
	
	public void addRelationshipBetweenActivities(String activity1, String activity2) {
		if (!indirectlyFollowsMatrix.containsKey(activity1)) {
			indirectlyFollowsMatrix.put(activity1, new HashMap<String, Integer>());
		}
		if (!maxSuccessOcc.containsKey(activity1)) {
			maxSuccessOcc.put(activity1, 0);
		}
		if (!invIndirectlyFollowsMatrix.containsKey(activity2)) {
			invIndirectlyFollowsMatrix.put(activity2, new HashMap<String, Integer>());
		}
		if (!maxPrecedOcc.containsKey(activity2)) {
			maxPrecedOcc.put(activity2, 0);
		}
		if (!indirectlyFollowsMatrix.get(activity1).containsKey(activity2)) {
			indirectlyFollowsMatrix.get(activity1).put(activity2, 0);
		}
		if (!invIndirectlyFollowsMatrix.get(activity2).containsKey(activity1)) {
			invIndirectlyFollowsMatrix.get(activity2).put(activity1, 0);
		}
		indirectlyFollowsMatrix.get(activity1).put(activity2, indirectlyFollowsMatrix.get(activity1).get(activity2) + 1);
		maxSuccessOcc.put(activity1, Math.max(maxSuccessOcc.get(activity1), indirectlyFollowsMatrix.get(activity1).get(activity2)));
		invIndirectlyFollowsMatrix.get(activity2).put(activity1, invIndirectlyFollowsMatrix.get(activity2).get(activity1) + 1);
		maxPrecedOcc.put(activity2, Math.max(maxPrecedOcc.get(activity2), invIndirectlyFollowsMatrix.get(activity2).get(activity1)));
	}
	
	public void clean(Double threshold) {
		List<String> indirectlyFollowsMatrixKeys = new ArrayList<String>(indirectlyFollowsMatrix.keySet());
		for (String key1 : indirectlyFollowsMatrixKeys) {
			Integer maxOcc = maxSuccessOcc.get(key1);
			List<String> subkeys = new ArrayList<String>(indirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				if (indirectlyFollowsMatrix.get(key1).get(key2) < (maxOcc * (1.0 - threshold))) {
					indirectlyFollowsMatrix.get(key1).remove(key2);
				}
			}
		}
		List<String> invIndirectlyFollowsMatrixKeys = new ArrayList<String>(invIndirectlyFollowsMatrix.keySet());
		for (String key1 : invIndirectlyFollowsMatrixKeys) {
			Integer maxOcc = maxPrecedOcc.get(key1);
			List<String> subkeys = new ArrayList<String>(invIndirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				if (invIndirectlyFollowsMatrix.get(key1).get(key2) < (maxOcc * (1.0 - threshold))) {
					invIndirectlyFollowsMatrix.get(key1).remove(key2);
				}
			}
		}
	}
	
	public void removeActivitiesBelowSupport(Integer minSupportForPlaces) {
		List<String> directlyFollowsMatrixKeys = new ArrayList<String>(indirectlyFollowsMatrix.keySet());
		for (String key1 : directlyFollowsMatrixKeys) {
			List<String> subkeys = new ArrayList<String>(indirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				if (indirectlyFollowsMatrix.get(key1).get(key2) < minSupportForPlaces) {
					indirectlyFollowsMatrix.get(key1).remove(key2);
				}
			}
		}
		List<String> invDirectlyFollowsMatrixKeys = new ArrayList<String>(invIndirectlyFollowsMatrix.keySet());
		for (String key1 : invDirectlyFollowsMatrixKeys) {
			List<String> subkeys = new ArrayList<String>(invIndirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				if (invIndirectlyFollowsMatrix.get(key1).get(key2) < minSupportForPlaces) {
					invIndirectlyFollowsMatrix.get(key1).remove(key2);
				}
			}
		}
	}
	
	public void calculateImportantRelationships(Map<String, Integer> activitiesCount, Double threshold) {
		List<String> indirectlyFollowsMatrixKeys = new ArrayList<String>(indirectlyFollowsMatrix.keySet());
		for (String key1 : indirectlyFollowsMatrixKeys) {
			Integer activityCount1 = activitiesCount.get(key1);
			
			if (!importantIndirectlyFollowsMatrix.containsKey(key1)) {
				importantIndirectlyFollowsMatrix.put(key1, new HashMap<String, Integer>());
			}
			
			List<String> subkeys = new ArrayList<String>(indirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				Integer activityCount2 = activitiesCount.get(key2);
				Double thisThreshold1 = activityCount1 * threshold;
				Double thisThreshold2 = activityCount2 * threshold;
				Integer value = indirectlyFollowsMatrix.get(key1).get(key2);
				boolean condition = (value > thisThreshold1 && value > thisThreshold2);
				
				if (condition) {
					importantIndirectlyFollowsMatrix.get(key1).put(key2, indirectlyFollowsMatrix.get(key1).get(key2));
				}
			}
		}
		List<String> invIndirectlyFollowsMatrixKeys = new ArrayList<String>(invIndirectlyFollowsMatrix.keySet());
		for (String key1 : invIndirectlyFollowsMatrixKeys) {
			Integer activityCount1 = activitiesCount.get(key1);
			
			if (!importantInvIndirectlyFollowsMatrix.containsKey(key1)) {
				importantInvIndirectlyFollowsMatrix.put(key1, new HashMap<String, Integer>());
			}
			
			List<String> subkeys = new ArrayList<String>(invIndirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				Integer activityCount2 = activitiesCount.get(key2);
				Double thisThreshold1 = activityCount1 * threshold;
				Double thisThreshold2 = activityCount2 * threshold;
				Integer value = invIndirectlyFollowsMatrix.get(key1).get(key2);
				boolean condition = (value > thisThreshold1 && value > thisThreshold2);

				//System.out.println("CONDITION = "+condition+" activity1="+key1+" activityCount1="+activityCount1+" activity2="+key2+" activityCount2="+activityCount2+" value="+value+" thisThreshold1="+thisThreshold1+" thisThreshold2="+thisThreshold2);

				if (condition) {
					importantInvIndirectlyFollowsMatrix.get(key1).put(key2, invIndirectlyFollowsMatrix.get(key1).get(key2));
				}
			}
		}
	}
	
	public void calculateReallyImportantRelationships() {
		List<String> indirectlyFollowsMatrixKeys = new ArrayList<String>(importantIndirectlyFollowsMatrix.keySet());
		for (String key1 : indirectlyFollowsMatrixKeys) {
			if (!reallyImportantIndirectlyFollowsMatrix.containsKey(key1)) {
				reallyImportantIndirectlyFollowsMatrix.put(key1, new HashMap<String, Integer>());
			}
			List<String> subkeys = new ArrayList<String>(importantIndirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				boolean isActivityReallyImportant = true;
				
				for (String key3 : subkeys) {
					if (!key3.equals(key2)) {
						boolean condition = importantIndirectlyFollowsMatrix.containsKey(key3) && importantIndirectlyFollowsMatrix.get(key3).containsKey(key2);
						
						if (condition) {
							isActivityReallyImportant = false;
							break;
						}
					}
				}
				
				if (isActivityReallyImportant) {
					reallyImportantIndirectlyFollowsMatrix.get(key1).put(key2, importantIndirectlyFollowsMatrix.get(key1).get(key2));
				}
			}
		}
		List<String> invIndirectlyFollowsMatrixKeys = new ArrayList<String>(importantInvIndirectlyFollowsMatrix.keySet());
		for (String key1 : invIndirectlyFollowsMatrixKeys) {
			if (!reallyImportantInvIndirectlyFollowsMatrix.containsKey(key1)) {
				reallyImportantInvIndirectlyFollowsMatrix.put(key1, new HashMap<String, Integer>());
			}
			List<String> subkeys = new ArrayList<String>(importantInvIndirectlyFollowsMatrix.get(key1).keySet());
			for (String key2 : subkeys) {
				if (!key1.equals(key2)) {
					boolean isActivityReallyImportant = true;
					
					for (String key3 : subkeys) {					
						if (!key3.equals(key2)) {
							boolean condition = importantInvIndirectlyFollowsMatrix.containsKey(key3) && importantInvIndirectlyFollowsMatrix.get(key3).containsKey(key2);
							//System.out.println("key1 = "+key1+" key2 = "+key2+" key3 = "+key3+" CONDITION="+condition);
	
							if (condition) {
								isActivityReallyImportant = false;
								break;
							}
						}
					}
					
					if (isActivityReallyImportant) {
						reallyImportantInvIndirectlyFollowsMatrix.get(key1).put(key2, importantInvIndirectlyFollowsMatrix.get(key1).get(key2));
					}
				}
			}
		}
	}
	
	public void printResume() {
		System.out.println("indirectlyFollowsMatrix=");
		System.out.println(indirectlyFollowsMatrix);
		System.out.println("maxSuccessOcc=");
		System.out.println(maxSuccessOcc);
		System.out.println("invIndirectlyFollowsMatrix=");
		System.out.println(invIndirectlyFollowsMatrix);
		System.out.println("maxPrecedOcc=");
		System.out.println(maxPrecedOcc);
		System.out.println("importantIndirectlyFollowsMatrix=");
		System.out.println(importantIndirectlyFollowsMatrix);
		System.out.println("importantInvIndirectlyFollowsMatrix=");
		System.out.println(importantInvIndirectlyFollowsMatrix);
		System.out.println("reallyImportantIndirectlyFollowsMatrix=");
		System.out.println(reallyImportantIndirectlyFollowsMatrix);
		System.out.println("reallyImportantInvIndirectlyFollowsMatrix=");
		System.out.println(reallyImportantInvIndirectlyFollowsMatrix);
	}
	
	public void addEvidentRelations(PlacesMemorizer memorizer, Set<String> possibleLeftActivitiesForPlacesAlgorithm, Set<String> possibleRightActivitiesForPlacesAlgorithm, Integer minSupportForPlaces, Map<String, Integer> activitiesCount) {
		Set<String> examinedActivitiesOnTheLeft = new HashSet<String>();
		Set<String> examinedActivitiesOnTheRight = new HashSet<String>();
		
		for (String key1 : reallyImportantIndirectlyFollowsMatrix.keySet()) {
			if (reallyImportantIndirectlyFollowsMatrix.get(key1).keySet().size() == 1) {
				String key2 = (new ArrayList<String>(reallyImportantIndirectlyFollowsMatrix.get(key1).keySet())).get(0);
				if (!key2.equals(key1)) {
					if (reallyImportantInvIndirectlyFollowsMatrix.containsKey(key2) && reallyImportantInvIndirectlyFollowsMatrix.get(key2).keySet().size() == 1) {
						String value = (new ArrayList<String>(reallyImportantInvIndirectlyFollowsMatrix.get(key2).keySet())).get(0);
						
						if (key1 == value) {
							if (possibleLeftActivitiesForPlacesAlgorithm.contains(key1) && possibleRightActivitiesForPlacesAlgorithm.contains(key2)) {
								Integer activityCount1 = activitiesCount.get(key1);
								Integer activityCount2 = activitiesCount.get(key2);
								boolean finalCondition = (activityCount1 > minSupportForPlaces && activityCount2 > minSupportForPlaces);
								
								System.out.println("minSupportForPlaces = "+minSupportForPlaces+" activityCount1 = "+activityCount1+" activityCount2 = "+activityCount2+" finalCondition="+finalCondition);
								
								if (finalCondition) {
									memorizer.addPlace(key1, key2);
									examinedActivitiesOnTheLeft.add(key1);
									examinedActivitiesOnTheRight.add(key2);
								}
							}
						}
					}
				}
			}
		}
	}
	
	public Map<OrderedActivity, Set<OrderedActivity>> getNonImportantRelationships(Set<String> activities, int direction, ActivitiesOccurrencesMemorizer activitiesOccurrences) {
		Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationships = new HashMap<OrderedActivity, Set<OrderedActivity>>();
		
		for (String activity0 : activities) {
			OrderedActivity activity = activitiesOccurrences.averageOrderedActivities.get(activity0);
			
			nonImportantRelationships.put(activity, new HashSet<OrderedActivity>());
			for (String otherActivity0 : activities) {
				OrderedActivity otherActivity = activitiesOccurrences.averageOrderedActivities.get(otherActivity0);
				if (!otherActivity0.equals(activity0)) {
					if ((importantIndirectlyFollowsMatrix.containsKey(activity0) && importantIndirectlyFollowsMatrix.get(activity0).containsKey(otherActivity0)) || (importantInvIndirectlyFollowsMatrix.containsKey(otherActivity) && importantInvIndirectlyFollowsMatrix.get(otherActivity).containsKey(activity))) {
						//System.out.println("NOT GOOD activity="+activity+" otherActivity="+otherActivity);
					}
					else {
						if (direction == 1) {
							if (indirectlyFollowsMatrix.containsKey(activity0) && indirectlyFollowsMatrix.containsKey(otherActivity0)) {
								Set<String> indFollowAct = new HashSet<String>(indirectlyFollowsMatrix.get(activity0).keySet());
								Set<String> indFollowOtherAct = new HashSet<String>(indirectlyFollowsMatrix.get(otherActivity0).keySet());
								indFollowAct.retainAll(indFollowOtherAct);
								if (indFollowAct.size() > 0) {
									nonImportantRelationships.get(activity).add(otherActivity);
								}
							}
						}
						else {
							if (invIndirectlyFollowsMatrix.containsKey(activity0) && invIndirectlyFollowsMatrix.containsKey(otherActivity0)) {
								Set<String> indFollowAct = new HashSet<String>(invIndirectlyFollowsMatrix.get(activity0).keySet());
								Set<String> indFollowOtherAct = new HashSet<String>(invIndirectlyFollowsMatrix.get(otherActivity0).keySet());
								indFollowAct.retainAll(indFollowOtherAct);
								if (indFollowAct.size() > 0) {
									nonImportantRelationships.get(activity).add(otherActivity);
								}
							}
						}
					}
				}
			}
		}
		
		return nonImportantRelationships;
	}
	
	public Map<OrderedActivity, Set<OrderedActivity>> getPossibleActivitiesOnTheRight(ActivitiesOccurrencesMemorizer activitiesOccurrences) {
		Map<OrderedActivity, Set<OrderedActivity>> possibleActivitiesOnTheRight = new HashMap<OrderedActivity, Set<OrderedActivity>>();
		
		for (String activity0 : indirectlyFollowsMatrix.keySet()) {
			OrderedActivity activity = activitiesOccurrences.averageOrderedActivities.get(activity0);
			for (String otherActivity0 : indirectlyFollowsMatrix.get(activity0).keySet()) {
				OrderedActivity otherActivity = activitiesOccurrences.averageOrderedActivities.get(otherActivity0);
				if (!otherActivity0.equals(activity0)) {
					if (!possibleActivitiesOnTheRight.containsKey(activity)) {
						possibleActivitiesOnTheRight.put(activity, new HashSet<OrderedActivity>());
					}
					possibleActivitiesOnTheRight.get(activity).add(otherActivity);
				}
			}
		}
		
		return possibleActivitiesOnTheRight;
	}
}
